import numpy as np

from pytest import mark

from ice.metrics.base import reduce_scores_dict
from ice.metrics.base import Sample
from ice.metrics.gold_paragraphs import get_gold_paragraph_df
from ice.metrics.nubia import Nubia
from ice.metrics.rouge import Rouge


def test_reduce_scores_to_single_value():
    scores = [dict(metric_a=0.1, metric_b=0.3), dict(metric_a=0.2, metric_b=0.4)]
    target_reduced_scores = dict(metric_a=0.15, metric_b=0.35)
    calculated_reduced_scores = reduce_scores_dict(scores, lambda x: sum(x) / len(x))
    for k, v in target_reduced_scores.items():
        assert k in calculated_reduced_scores
        assert np.isclose(calculated_reduced_scores[k], v)


@mark.anyio
async def test_rouge_basic_precision():
    samples = [
        Sample(left=["I can haz cheeseburger"], right=["I make cats haz cheeseburger"])
    ]
    scores = await Rouge().compute(samples)

    # Basic sanity check
    rouge_1_p = scores[0].dict()["rouge_1"]["p"]
    rouge_2_p = scores[0].dict()["rouge_2"]["p"]

    assert rouge_1_p == 3 / 4
    assert rouge_2_p == 1 / 3


@mark.anyio
@mark.parametrize(
    "left,right,expected_result",
    [
        (["subsequence"], ["sequence including subsequence"], 1 / 3),
        (["sequence including subsequence"], ["subsequence"], 1),
        (["sequence including subsequence"], ["sequence", "other random crap"], 0.25),
    ],
)
async def test_rouge_recall(left, right, expected_result):
    samples = [Sample(left=left, right=right)]
    scores = await Rouge().compute(samples)

    rouge_l_r = scores[0].dict()["rouge_l"]["r"]

    assert rouge_l_r == expected_result


@mark.anyio
async def test_rouge_multiple_hypotheses_and_references_matches_single_score_when_inputs_are_identical():
    sample_single = [
        Sample(left=["I can haz cheeseburger"], right=["I make cats haz cheeseburger"])
    ]
    scores_single = await Rouge().compute(sample_single)

    samples_repeated = [
        Sample(
            left=["I can haz cheeseburger", "I can haz cheeseburger"],
            right=["I make cats haz cheeseburger", "I make cats haz cheeseburger"],
        )
    ]
    scores_repeated = await Rouge().compute(samples_repeated)

    assert scores_single == scores_repeated


@mark.skip(reason="Requires OUGHT_INFERENCE_API_KEY")
@mark.slow
@mark.anyio
async def test_nubia():
    m = Nubia()
    sample_single = [
        Sample(left=["I can haz cheeseburger"], right=["I cannot haz cheeseburger"])
    ]
    resp = await m.compute(sample_single)
    assert len(resp) == 1 and len(resp[0]) == 1
    # These are strongly contradictory & the model is deterministic
    assert resp[0][0].contradiction > 0.99


@mark.skip(reason="Parses all PDFs - very slow")
@mark.slow
@mark.anyio
async def test_gold_paragraphs():
    gold_paragraphs_df, id_to_paragraph = get_gold_paragraph_df("placebo")

    assert "quote" in gold_paragraphs_df.columns
    assert "paragraph" in gold_paragraphs_df.columns
    assert "paragraph_id" in gold_paragraphs_df.columns

    assert len(gold_paragraphs_df) > 20

    for _, paragraph_id in gold_paragraphs_df["paragraph_id"].iteritems():
        paragraph = id_to_paragraph[paragraph_id]
        assert not paragraph.is_empty()



from collections.abc import Mapping
from collections.abc import Sequence
from typing import cast

import numpy as np

from anyio import ExceptionGroup
from structlog.stdlib import get_logger

from ice.agents.ought_inference import OughtInferenceAgent
from ice.apis.openai import openai_complete
from ice.apis.openai import TooLongRequestError
from ice.paper import Paper
from ice.recipe import recipe
from ice.recipes.best_completion import best_completion
from ice.recipes.best_completion import completion_perplexity
from ice.recipes.consort_flow import baseline_elicit_answer
from ice.recipes.find_best_few_shot_prompt import score_few_shot
from ice.recipes.meta.eval_paper_qa.types import PaperQaGoldStandard
from ice.recipes.program_search.nodes.answer.types import Demonstration
from ice.recipes.program_search.nodes.prune.prune import prune
from ice.recipes.program_search.nodes.select.prompts import get_selections
from ice.recipes.program_search.nodes.select.prompts import make_selection_prompt
from ice.recipes.program_search.nodes.select.prompts import RenderableSelectionExample
from ice.recipes.program_search.types import (
    remove_highest_perplexity,
)
from ice.utils import map_async
from ice.utils import n_tokens
from ice.utils import reduce_async
from ice.utils import window_dropping

log = get_logger()


def last_token_logprob(openai_response: dict) -> float:
    return openai_response["choices"][0]["logprobs"]["token_logprobs"][-1]


def last_token_top_logprobs(openai_response: dict) -> dict[str, float]:
    return openai_response["choices"][0]["logprobs"]["top_logprobs"][-1]


def logprobs_greater_than_none(
    selections: Mapping[int, float], none_logprob: float, texts: Sequence[str]
) -> Sequence[str]:
    return [text for idx, text in enumerate(texts) if selections[idx] > none_logprob]


async def select(
    question: str,
    texts: Sequence[str],
    existing: Sequence[str],
    examples: list[RenderableSelectionExample] | None = None,
) -> Sequence[str]:
    """Select additional texts by comparing logprobs of indices considered by the model.

    Args:
        question (str): The question to select texts for.
        texts (Sequence[str]): Texts to consider for selection.
        existing (Sequence[str]): Already-selected texts

    Returns:
        Sequence[str]: Newly selected texts (subset of texts)
    """
    if len(texts) > 5:
        log.warning(
            "The OpenAI API only returns the top 5 logprobs, so passing more than 5 candidates means that not all can be fully considered.",
            num_candidates=len(texts),
        )
    prompt = make_selection_prompt(
        question=question,
        existing=existing,
        texts=[t for t in texts if t],
        examples=examples,
    )
    try:
        response = await openai_complete(
            prompt=prompt, max_tokens=0, logprobs=100, echo=True
        )
    except TooLongRequestError:
        if examples and len(examples) >= 2:
            return await select(question, texts, existing, examples[:-1])
        else:
            raise
    choice_logprobs = get_selections(last_token_top_logprobs(response), len(texts))
    return logprobs_greater_than_none(
        choice_logprobs, last_token_logprob(response), texts
    )


async def maybe_binary_prune(question: str, existing: list[str], max_to_keep=8):
    try:
        return await prune(question, existing, max_to_keep=8)
    except TooLongRequestError:
        mid = len(existing) // 2
        h1 = await maybe_binary_prune(question, existing[:mid], max_to_keep=max_to_keep)
        h2 = await maybe_binary_prune(question, existing[mid:], max_to_keep=max_to_keep)
        return await maybe_binary_prune(question, h1 + h2, max_to_keep=max_to_keep)


async def select_reduce(
    question: str,
    texts: Sequence[Sequence[str]],
    do_prune: bool = False,
    examples: list[RenderableSelectionExample] | None = None,
) -> Sequence[str]:
    """Select texts that answer the question by reducing over `select`

    Args:
        question (str): The question to select texts for.
        texts (Sequence[Sequence[str]]): Texts to consider for selection, split into groups to consider at each step.

    Returns:
        Sequence[str]: Selected texts.
    """

    async def select_some(existing: list[str], new_texts: Sequence[str]):
        try:
            new_selections = await select(question, new_texts, existing, examples)
        except TooLongRequestError:
            if do_prune:
                existing = await maybe_binary_prune(
                    question, existing, max_to_keep=8
                )  # TODO: Be smarter about the limit here
                return await select_some(existing, new_texts)
            else:
                log.warning("Skipping because prompt full")
            return existing
        return existing + list(new_selections)

    return await reduce_async(select_some, texts, cast(list[str], []))


async def windowed_select(
    question: str,
    texts: Sequence[str],
    n: int,
    step: int,
    examples: list[RenderableSelectionExample] | None = None,
) -> Sequence[bool]:
    """Select texts that answer the question via

    Args:
        question (str): The question to select texts for.
        texts (Sequence[str]): Texts to consider for selection.
        n (int): Number of texts to consider at each step.
        step (int): Overlap between windows. (if n == step, partition the document; if step < n, window with step size).

    Returns:
        Sequence[str]: Selected texts.
    """
    windowed_texts = window_dropping(texts, n, step)
    selections = set(
        await select_reduce(question, windowed_texts, do_prune=True, examples=examples)
    )
    return [t in selections for t in texts]


async def select_results_using_elicit_prompt(  # Best recall [use this]
    question: str,
    texts: Sequence[str],
) -> Sequence[tuple[str, float]]:
    """Select texts that answer the question via perplexity of the "not mentioned"
    response to an elicit-like question-answering prompt
    """

    prompts = [
        baseline_elicit_answer.elicit_qa_prompt(
            qa_question=question,
            excerpt=text,
        )
        for text in texts
    ]

    completion = " " + baseline_elicit_answer.COMBINED_NA_PHRASE

    prompt_perplexities = await best_completion(
        prompts=prompts,
        completion=completion,
    )

    return [(t, p[1]) for t, p in zip(texts, prompt_perplexities)]
    # Lower perplexity means more likely to be "not mentioned in excerpt"


def filter_by_perplexity_threshold(
    results: Sequence[tuple[str, float]], threshold: float
):
    return [r for r in results if r[1] > threshold]


def to_paragraphs(paper: Paper) -> Sequence[str]:
    return [str(p) for p in paper.nonempty_paragraphs()]


async def elicit_negative_few_shot_example(
    example: PaperQaGoldStandard, threshold: float = 1.16, max_examples: int | None = 4
) -> Demonstration | None:
    paragraphs = to_paragraphs(example.paper)
    gold_support = set(example.gold_support)
    assert gold_support.issubset(
        set(paragraphs)
    ), "Expected gold support to already be in paragraph form"
    search_results = await select_results_using_elicit_prompt(
        question=example.question, texts=paragraphs
    )
    most_relevant = filter_by_perplexity_threshold(search_results, threshold=threshold)
    most_relevant_not_actually_relevant = [
        t for t in most_relevant if t[0] not in gold_support
    ]
    if not most_relevant_not_actually_relevant:
        return None
    if max_examples:
        while len(most_relevant_not_actually_relevant) > max_examples:
            scored_support = await _score_support(
                question=example.question,
                answer=example.short_gold_answer,
                support_candidates=[t[0] for t in most_relevant_not_actually_relevant],
            )
            most_relevant_not_actually_relevant = remove_highest_perplexity(
                scored_support
            )

    return baseline_elicit_answer.convert_to_non_answer(
        Demonstration(
            question=example.question,
            texts=[t[0] for t in most_relevant_not_actually_relevant],
            answer="",
        )
    )


async def _score_support(
    question: str, answer: str, support_candidates: Sequence[str]
) -> Sequence[tuple[str, float]]:
    async def get_perplexity(support: str) -> tuple[str, float]:
        return support, await completion_perplexity(
            prompt=baseline_elicit_answer.elicit_qa_prompt(
                qa_question=question, excerpt=support
            ),
            completion=" " + answer.strip(),
        )

    perplexities = await map_async(support_candidates, get_perplexity)
    # Lower is better in this case
    return perplexities


async def positive_few_shot_example(
    example: PaperQaGoldStandard,
    threshold: float = 1.16,
    max_examples: int | None = 4,
    add_noise: bool = False,
) -> Demonstration:
    paragraphs = to_paragraphs(example.paper)
    gold_support = set(example.gold_support)
    assert gold_support.issubset(
        set(paragraphs)
    ), "Expected gold support to already be in paragraph form"
    search_results = await select_results_using_elicit_prompt(
        question=example.question, texts=paragraphs
    )
    most_relevant_all = filter_by_perplexity_threshold(
        search_results, threshold=threshold
    )
    most_relevant = most_relevant_all
    most_relevant_texts = {t[0] for t in most_relevant}
    noisy_support = [
        p
        for p in paragraphs
        if p in most_relevant_texts and add_noise or p in gold_support
    ]

    while max_examples and most_relevant and len(noisy_support) > max_examples:
        scored_noisy_support = await _score_support(
            question=example.question,
            answer=example.short_gold_answer,
            support_candidates=noisy_support,
        )
        most_relevant = remove_highest_perplexity(scored_noisy_support)
        most_relevant_texts = {t[0] for t in most_relevant}
        noisy_support = [p for p in paragraphs if p in most_relevant_texts]

    return Demonstration(
        question=example.question, texts=noisy_support, answer=example.short_gold_answer
    )


async def select_using_elicit_prompt_few_shot(
    question: str,
    texts: Sequence[str],
    examples: Sequence[PaperQaGoldStandard],
    n_shots: int = 5,
    seed: int = 0,
    include_negative: bool = False,
) -> Sequence[tuple[str, float]]:
    _examples = list(examples)
    rng = np.random.default_rng(seed)
    rng.shuffle(_examples)  # type: ignore[arg-type]

    example_separator = "\n\n---\n\n"

    if include_negative:
        demonstrations_or_none = [
            (await elicit_negative_few_shot_example(example, max_examples=1))
            if idx % 3 == 0  # more positive than negative examples
            else (await positive_few_shot_example(example, max_examples=1))
            for idx, example in enumerate(examples)
        ]
    else:
        demonstrations_or_none = [
            await positive_few_shot_example(example, max_examples=1)
            for example in examples
        ]
    demonstrations = [d for d in demonstrations_or_none if d is not None]

    prompts_and_completions = baseline_elicit_answer.make_few_shot_examples(
        demonstrations
    )

    try:
        few_shot_prompts = await score_few_shot(
            examples_prompts=[pc[0] for pc in prompts_and_completions],
            examples_completions=[pc[1] for pc in prompts_and_completions],
            n_shots=n_shots,
            split_string=example_separator,
            prefix="",
            max_permutations=10,  # TODO: tune these a bit
            max_test_size=5,
            seed=seed,
        )
    except ExceptionGroup as es:
        if not all((isinstance(e, TooLongRequestError) for e in es.exceptions)):
            raise
        if n_shots:
            return await select_using_elicit_prompt_few_shot(
                question=question,
                texts=texts,
                examples=examples,
                n_shots=n_shots - 1,
                seed=seed,
            )
        else:
            raise
    except TooLongRequestError:
        if n_shots:
            return await select_using_elicit_prompt_few_shot(
                question=question,
                texts=texts,
                examples=examples,
                n_shots=n_shots - 1,
                seed=seed,
            )
        else:
            raise

    best_few_shot_prompt = min(few_shot_prompts, key=lambda p: p[1])[0]

    prompts = [
        best_few_shot_prompt
        + example_separator
        + baseline_elicit_answer.elicit_qa_prompt(
            qa_question=question,
            excerpt=text,
        )
        for text in texts
    ]

    completion = " " + baseline_elicit_answer.COMBINED_NA_PHRASE

    if any((n_tokens(prompt + " " + completion) > 4_095 for prompt in prompts)):
        if not n_shots:
            raise ValueError("Prompt would be too long, cannot shorten")
        return await select_using_elicit_prompt_few_shot(
            question=question,
            texts=texts,
            examples=examples,
            n_shots=n_shots - 1,
            seed=seed,
        )

    prompt_perplexities = await best_completion(
        prompts=prompts,
        completion=completion,
    )

    return [
        (text, prompt_perplexity[1])
        for text, prompt_perplexity in zip(texts, prompt_perplexities)
    ]
    # Lower perplexity means more likely to be "not mentioned in excerpt"


async def select_results_using_top_monot5_paragraph(
    question: str,
    texts: Sequence[str],
) -> str:
    agent = OughtInferenceAgent(engine="mono-t5-base-qa")

    scores = await agent.relevance_batch(
        question=question,
        contexts=texts,
    )

    assert len(scores) == len(texts)

    return max(zip(texts, scores), key=lambda t: t[1])[0]


def as_strings(selections: Sequence[bool], texts: Sequence[str]) -> Sequence[str]:
    return [t for t, s in zip(texts, selections) if s]


def as_bool(selections: Sequence[str], texts: Sequence[str]) -> Sequence[bool]:
    selections_set = set(selections)
    return [t in selections_set for t in texts]


recipe.main(windowed_select)



import httpx

from fvalues import F

from ice.recipe import recipe


def make_qa_prompt(context: str, question: str) -> str:
    return F(
        f"""
Background text: "{context}"

Answer the following question about the background text above:

Question: "{question}"
Answer: "
"""
    ).strip()


async def search(query: str = "Who is the president of the United States?") -> dict:
    async with httpx.AsyncClient() as client:
        params = {"q": query, "hl": "en", "gl": "us", "api_key": "e29...b4c"}
        response = await client.get("https://serpapi.com/search", params=params)
        return response.json()


recipe.main(search)



from fvalues import F

from ice.recipe import recipe
from ice.recipes.primer.subquestions import ask_subquestions
from ice.utils import map_async


def make_qa_prompt(question: str) -> str:
    return F(
        f"""Answer the following question:

Question: "{question}"
Answer: "
"""
    ).strip()


async def answer(question: str) -> str:
    prompt = make_qa_prompt(question)
    answer = await recipe.agent().complete(prompt=prompt, stop='"')
    return answer


async def answer_by_amplification(
    question: str = "What is the effect of creatine on cognition?",
):
    subquestions = await ask_subquestions(question=question)
    subanswers = await map_async(subquestions, answer)
    return list(zip(subquestions, subanswers))


recipe.main(answer_by_amplification)



import functools
import re
import time

from enum import IntEnum
from typing import Tuple

import torch

from structlog.stdlib import get_logger
from transformers.models.auto.modeling_auto import AutoModelForSeq2SeqLM
from transformers.models.auto.tokenization_auto import AutoTokenizer
from transformers.models.bart.modeling_bart import BartForConditionalGeneration
from transformers.models.bart.modeling_bart import BartForSequenceClassification
from transformers.models.bart.tokenization_bart_fast import BartTokenizerFast
from transformers.models.t5.modeling_t5 import T5ForConditionalGeneration
from transformers.models.t5.tokenization_t5_fast import T5TokenizerFast

from ice.paper import split_sentences

log = get_logger()


@functools.cache
def get_device() -> torch.device:
    device_name = "cuda" if torch.cuda.is_available() else "cpu"
    if device_name == "cpu":
        log.warning("Running on CPU could be *very* slow.")
    log.info(f"Running on {str(device_name).upper()}.")
    return torch.device(device_name)


@functools.cache
def load_T5() -> Tuple[T5TokenizerFast, T5ForConditionalGeneration]:
    device = get_device()
    T5_tokenizer = AutoTokenizer.from_pretrained("castorini/monot5-base-msmarco-10k")

    T5_model = AutoModelForSeq2SeqLM.from_pretrained(
        "castorini/monot5-base-msmarco-10k"
    ).to(device)
    return T5_tokenizer, T5_model


@functools.cache
def load_BART() -> Tuple[BartTokenizerFast, BartForConditionalGeneration]:
    device = get_device()
    BART_tokenizer = BartTokenizerFast.from_pretrained("facebook/bart-large-mnli")

    BART_model = BartForSequenceClassification.from_pretrained(
        "facebook/bart-large-mnli"
    ).to(device)
    return BART_tokenizer, BART_model


@functools.cache
def load_T0(
    model_name="bigscience/T0_3B",
) -> Tuple[T5TokenizerFast, T5ForConditionalGeneration]:
    # For testing
    model_name = "t5-small"
    device = get_device()
    T0_tokenizer = AutoTokenizer.from_pretrained(model_name)

    i: float = time.time()
    T0_model = AutoModelForSeq2SeqLM.from_pretrained(model_name).to(device)
    log.info(f"Loaded T0 in {time.time() - i} seconds.")
    return T0_tokenizer, T0_model


def get_sentence(paragraph: str, number: str) -> str:
    for sentence in split_sentences(paragraph):
        if number in sentence:
            return sentence
    return ""


class IsCategory(IntEnum):
    NO = 0
    YES = 1


def BART_classify(
    category: str, paragraph: str, number: str, BART_threshold=2 / 3
) -> IsCategory:  # Classification is either "Y" or "N"
    device = get_device()
    BART_tokenizer, BART_model = load_BART()
    hypothesis = f"The number {number} is about the {category} of the study."
    premise = get_sentence(paragraph, number)
    if not premise:
        log.info(f"Could not find number {number} in paragraph")
        return IsCategory.NO
    input_ids = BART_tokenizer.encode(
        premise, hypothesis, return_tensors="pt", max_length=512, truncation=True
    ).to(device)
    with torch.no_grad():
        logits = BART_model(input_ids)[0][:, [0, 2]]
        logits = logits.softmax(dim=-1)[0]
    return IsCategory((logits[1] > BART_threshold).item())


def T5_predict(question: str, document: str) -> float:
    device = get_device()
    T5_tokenizer, T5_model = load_T5()
    T5_input = f"Query: {question} Document: {document} Relevant:"
    inputs = T5_tokenizer.encode(
        T5_input, return_tensors="pt", max_length=512, truncation=True
    ).to(device)
    false_token_index = T5_tokenizer.encode("false", return_tensors="pt")[0][0].to(
        device
    )
    true_token_index = T5_tokenizer.encode("true", return_tensors="pt")[0][0].to(device)
    outputs = T5_model.generate(
        inputs, return_dict_in_generate=True, output_scores=True
    )
    scores = outputs["scores"][0][:, [true_token_index, false_token_index]].softmax(
        dim=-1
    )
    return scores[0][0].item()


def T5_score(question: str, document: str) -> float:
    scores = []
    for sentence in split_sentences(document):
        scores.append(T5_predict(question, sentence))
    if not scores:
        return 0
    return sum(scores) / len(
        scores
    )  # Could use max(scores) instead of sum(scores)/len(scores)


def T0_classify(
    category: str, abstract: str, paragraph: str, number: str, t0_threshold=0.45
) -> IsCategory:
    device = get_device()
    T0_tokenizer, T0_model = load_T0()
    question = f"Is the number {number} is about the {category} of the study?"
    premise = get_sentence(paragraph, number)
    if not premise:
        log.info(f"Could not find number {number} in paragraph")
        return IsCategory.NO
    context = abstract + premise
    prompt = f"""{context}
Given the context: {question}
Possible answers: Yes the number {number} is about the {category} of the study?, No the number {number} is not about the {category} of the study?"""

    # Run T0_3B
    inputs = T0_tokenizer.encode(
        prompt, return_tensors="pt", max_length=1024, truncation=True
    ).to(device)
    outputs = T0_model.generate(
        inputs, return_dict_in_generate=True, output_scores=True, max_new_tokens=2
    )
    false_token_index, true_token_index = (
        T0_tokenizer.encode("No")[0],
        T0_tokenizer.encode("Yes")[0],
    )
    scores = outputs["scores"][0][:, [true_token_index, false_token_index]].softmax(
        dim=-1
    )
    scores = scores[0]

    return IsCategory((scores[0] > t0_threshold).item())


# Credit: https://stackoverflow.com/questions/39936527/python-removing-references-from-a-scientific-paper

# Remove citations
def remove_citations(s: str) -> str:
    return re.sub(r"\s\([A-Z][a-z]+,\s[A-Z][a-z]?\.[^\)]*,\s\d{4}\)", "", s)


threshold = 0.0009468319839808456
# Max score threshold is 0.001010593439036711
# Average score threshold is 0.0009468319839808456


def BERT_T5_T0(
    category: str,
    abstract: str,
    paragraph: str,
    number: str,
    threshold=0.0010661712990630886,
) -> IsCategory:
    paragraph, abstract = remove_citations(paragraph), remove_citations(abstract)
    score = T5_score(f"What is the {category} of this study?", paragraph)
    if score < threshold:
        log.info("T5 score too low", t5_score=score)
        return IsCategory.NO
    bert_classification = BART_classify(category, paragraph, number)
    if bert_classification == IsCategory.NO:
        log.info("Failed BART")
        return IsCategory.NO
    T0_classification = T0_classify(category, abstract, paragraph, number)
    if T0_classification == IsCategory.NO:
        log.info("Failed T0")
        return IsCategory.NO
    log.info(
        "BERT, T5 and T0 agree on the answer Y.", paragraph=paragraph, number=number
    )
    return IsCategory.YES



import os
import subprocess
import sys
import time

import httpx

from structlog import get_logger

from ice.settings import server_url
from ice.settings import settings

log = get_logger()


def is_server_running():
    from ice.routes.app import PING_RESPONSE

    try:
        response = httpx.get(server_url() + "/ping")
        return response.text == PING_RESPONSE
    except httpx.HTTPError:
        return False


def wait_until_server_running():
    start_time = time.time()
    while not is_server_running():
        if time.time() - start_time > 10:
            raise TimeoutError("Server didn't start within 10 seconds")
        time.sleep(0.1)


def ensure_server_running():
    if is_server_running():
        return

    log.info("Starting server, set OUGHT_ICE_AUTO_SERVER=0 to disable.")
    subprocess.Popen(
        [sys.executable, "-m", "ice.server", "start"],
        env=os.environ,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    wait_until_server_running()
    log.info("Server started! Run `python -m ice.server stop` to stop it.")


def stop():
    if not is_server_running():
        print(
            f"Server at {server_url()} is not running, maybe set OUGHT_ICE_HOST/PORT?"
        )
        sys.exit(0)

    url = server_url() + "/stop"
    print(f"Posting to {url}...")
    try:
        httpx.post(url)
    except httpx.RemoteProtocolError:
        pass  # expected as the server dies

    if is_server_running():
        print("Weird, the server is still running!")
        sys.exit(1)
    else:
        print("Server stopped!")


def start():
    import uvicorn

    params_by_name = {p.name: p for p in uvicorn.main.params}
    params_by_name["app"].default = "ice.routes.app:app"
    params_by_name["host"].default = settings.OUGHT_ICE_HOST
    params_by_name["port"].default = settings.OUGHT_ICE_PORT

    uvicorn.main()


def main():
    if sys.argv[1:2] == ["stop"]:
        if sys.argv[2:]:
            print(
                "The stop command takes no arguments, but you can set OUGHT_ICE_HOST/PORT environment variables."
            )
            sys.exit(1)
        stop()
        sys.exit(0)

    if sys.argv[1:2] == ["start"]:
        del sys.argv[1]  # remove redundant command

    start()


if __name__ == "__main__":
    main()



from urllib.parse import urljoin

import httpx

from tenacity import retry
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_random_exponential

from ice.agents.base import Agent
from ice.settings import settings


class OughtInferenceAgent(Agent):
    def __init__(self, engine: str):
        self.url = urljoin(settings.OUGHT_INFERENCE_URL, engine)

    @retry(wait=wait_random_exponential(), stop=stop_after_attempt(3))
    async def relevance(
        self, *, question, context, verbose=False, default=None
    ) -> float:
        async with httpx.AsyncClient() as client:
            client.headers["x-api-key"] = settings.OUGHT_INFERENCE_API_KEY
            response = await client.post(
                self.url, json=dict(query=question, documents=[context])
            )
            response.raise_for_status()
        return response.json()["results"][0]["score"]

    @retry(wait=wait_random_exponential(), stop=stop_after_attempt(3))
    async def relevance_batch(
        self, *, question, contexts, verbose=False, default=None
    ) -> list[float]:
        async with httpx.AsyncClient() as client:
            client.headers["x-api-key"] = settings.OUGHT_INFERENCE_API_KEY
            response = await client.post(
                self.url, json=dict(query=question, documents=contexts)
            )
            response.raise_for_status()
        return [r["score"] for r in response.json()["results"]]



"""
Decorator for caching function results to disk
"""
import asyncio
import functools
import inspect

from pathlib import Path

from ice.settings import CACHE_DIR
from ice.sqlite_shelf import SQLiteShelf


def diskcache(cache_dir: Path = CACHE_DIR):
    def get_cache(func):
        cache_file = (cache_dir / func.__name__).as_posix() + ".sqlite"
        return SQLiteShelf(cache_file, "diskcache")

    def get_cache_key(func, *args, **kwargs):
        return repr(inspect.getcallargs(func, *args, **kwargs))

    def decorator(func):
        @functools.wraps(func)
        def sync_wrapper(*args, **kwargs):
            key = get_cache_key(func, *args, **kwargs)
            cache = get_cache(func)
            if key in cache:
                return cache[key]
            result = func(*args, **kwargs)
            cache[key] = result
            return result

        @functools.wraps(func)
        async def async_wrapper(*args, **kwargs):
            key = get_cache_key(func, *args, **kwargs)
            cache = get_cache(func)
            if key in cache:
                return cache[key]
            cache.close()
            # close connection while waiting for async function so we don't have a bunch of connections open simultaneously

            result = await func(*args, **kwargs)

            cache = get_cache(func)
            cache[key] = result
            return result

        if asyncio.iscoroutinefunction(func):
            return async_wrapper

        return sync_wrapper

    return decorator



from collections.abc import Mapping

from numerizer.consts import ALL_ORDINALS

from ice.formatter.transform import _Transform


class PositionalTransform(_Transform):
    """
    Implement to convert the index of the example to text.
    """

    def transform(self, position: int, total: int) -> str:
        raise NotImplementedError


class OrdinalWord(PositionalTransform):
    """
    Complete with the English ordinal ("first", "second", etc.) up to
    "twentieth", based on the position of the example within the prompt.

    Optionally, special case the last position (e.g., use "finally" or "last" instead
    of an ordinal), and override any other cases by providing a `special` mapping.
    """

    def __init__(
        self,
        capitalize=False,
        finally_case: str | None = None,
        special: Mapping[int, str] | None = None,
    ):
        self.capitalize = capitalize
        self.finally_case = finally_case
        self.words = {(int(v) - 1): k for k, v in ALL_ORDINALS.items()}
        if special:
            for position, word in special.items():
                self.words[position] = word

    def transform(self, position: int, total: int) -> str:
        if position == total - 1 and self.finally_case is not None:
            word = self.finally_case
        else:
            try:
                word = self.words[position]
            except KeyError:
                raise ValueError(f"Position {position} not in ordinal word dictionary")
        return word.capitalize() if self.capitalize else word



from collections import Counter
from collections.abc import Sequence

from typing_extensions import assert_never

from ice.formatter.multi import stop
from ice.formatter.multi import StopSentinel
from ice.formatter.transform.value import ValueTransform
from ice.recipes.experiments_and_arms.types import PassageWithReasoning
from ice.recipes.experiments_and_arms.types import ReasoningStage


def start_last_example(
    helpfulness: str | None,
    reasoning: str | None,
    pre_final: str | None = None,
    pre_helpful: str | None = None,
    pre_answer: str | None = None,
) -> dict[str, ValueTransform[Sequence[str]] | str | StopSentinel]:
    assert (
        not helpfulness or reasoning
    ), "Final reasoning required alongside helpfulness"

    state: ReasoningStage = (
        "reasoning" if not reasoning else "helpfulness" if not helpfulness else "answer"
    )

    match state:
        case "reasoning":
            return {"reasoning": stop(pre_final or "")}
        case "helpfulness":
            assert reasoning
            return {
                "reasoning": reasoning,
                "helpfulness": stop(pre_helpful or ""),
            }
        case "answer":
            assert reasoning and helpfulness
            return {
                "reasoning": reasoning,
                "helpfulness": helpfulness,
                "answer": stop(pre_answer or ""),
            }
        case _:
            assert_never(state)


def get_part(response: str, pre_part: str, post_part: str) -> str:
    split = response.split(pre_part)
    remaining = split[1] if len(split) > 1 else split[0]
    return remaining.split(post_part)[0].strip()


async def plurality_greedy(
    samples: Sequence[PassageWithReasoning[int]],
) -> PassageWithReasoning[int]:
    most_common = Counter(
        [s.final_answer for s in samples if s.final_answer is not None]
    ).most_common(1)[0][0]
    for sample in samples:
        if sample.final_answer == most_common:
            return sample
    raise RuntimeError("Unreachable")



from fvalues import F

from ice.recipes.primer.debate.types import *


def initialize_debate(question: Message) -> Debate:
    return [
        ("Question", question),
        ("Alice", "I'm in favor."),
        ("Bob", "I'm against."),
    ]


def render_debate(debate: Debate, self_name: Name | None = None) -> str:
    debate_text = ""
    for speaker, text in debate:
        if speaker == self_name:
            speaker = "You"
        debate_text += F(f'{speaker}: "{text}"\n')
    return debate_text.strip()



from collections.abc import Callable
from collections.abc import Sequence
from typing import TypeVar

from structlog.stdlib import get_logger

from ice.apis.openai import TooLongRequestError
from ice.recipe import recipe
from ice.recipes.experiments_and_arms.recipes.reason_select_and_answer import (
    answer_with_best_reasoning,
)
from ice.recipes.experiments_and_arms.types import MultipartReasoningPrompt
from ice.recipes.experiments_and_arms.types import PassageWithReasoning
from ice.utils import map_async
from ice.utils import window_dropping

log = get_logger()

T = TypeVar("T")


def choices_log_probs(response: dict, choices: Sequence[str]):
    top_logprobs: dict[str, float] = response["choices"][0]["logprobs"]["top_logprobs"][
        0
    ]
    return {choice: top_logprobs.get(choice) for choice in choices}


async def rank_passages_selector(
    samples: Sequence[PassageWithReasoning[float]],
) -> PassageWithReasoning[float]:
    """Select the reasoning with the final answer closest to the mean.

    Args:
        samples (Sequence[PassageWithReasoning[float]]): Passages with reasoning to rank.

    Returns:
        PassageWithReasoning[float]: Passage with reasoning closest to the mean.
    """
    scores = [sample.final_answer for sample in samples if sample.final_answer]
    mean_score = (
        sum(scores) / len(scores) if scores else -100_000
    )  # really small non-infinite number
    closest = min(
        samples,
        key=lambda sample: abs(sample.final_answer - mean_score)
        if sample.final_answer
        else float("inf"),
    )
    return PassageWithReasoning(
        passage=closest.passage,
        reasoning=closest.reasoning,
        helpfulness=closest.helpfulness,
        score=mean_score,
        final_answer=mean_score,
    )


async def initial_passages(
    passages: Sequence[str],
    prompt_func: Callable[[int], MultipartReasoningPrompt],
    choices: Sequence[str],
    best_choice: str,
    *,
    num_samples: int = 3,
    reasoning_stop: tuple[str, ...],
    get_reasoning: Callable[[str], str],
    get_helpfulness: Callable[[str], str],
    num_shots: int,
    passages_per_prompt: int = 4,
    step: int = 1,
):
    # Dummy recipe to experiment with just returning the initial part of the doc
    candidates = window_dropping(items=passages, n=passages_per_prompt, step=step)[0]
    sorted_answers = await rate_helpfulness_with_reasoning(
        passages=candidates,
        prompt_func=prompt_func,
        choices=choices,
        best_choice=best_choice,
        num_samples=num_samples,
        reasoning_stop=reasoning_stop,
        get_reasoning=get_reasoning,
        get_helpfulness=get_helpfulness,
        num_shots=num_shots,
        passages_per_prompt=passages_per_prompt,
        step=step,
    )
    return sorted_answers


async def rate_helpfulness_with_reasoning(
    passages: Sequence[str],
    prompt_func: Callable[[int], MultipartReasoningPrompt],
    choices: Sequence[str],
    best_choice: str,
    *,
    num_samples: int = 1,
    reasoning_stop: tuple[str, ...],
    get_reasoning: Callable[[str], str],
    get_helpfulness: Callable[[str], str],
    num_shots: int,
    passages_per_prompt: int = 4,
    step: int = 1,
) -> Sequence[PassageWithReasoning[float]]:
    """Rank passages by final answer probability after sampling reasoning.

    Args:
        passages (Sequence[str]): Paragraphs to break into longer passages and rank
        prompt_func (Callable[[int], MultipartReasoningPrompt]): Function that returns a multipart reasoning prompt with the given number of examples
        choices (Sequence[str]): Final answer choices
        best_choice (str): Choice in choices that should be considered "correct" for purposes of ranking by logprobs
        reasoning_stop (tuple[str, ...]): Stop sequence(s) for reasoning prompt
        get_reasoning (Callable[[str], str]): Extract reasoning from a completion
        get_helpfulness (Callable[[str], str]): Extract helpfulness statement from a completion
        num_shots (int): Number of few-shot examples to use in prompts
        num_samples (int, optional): Number of samples for each passage; if >1, scores are averaged and the final reasoning closest to the average is chosen. Defaults to 1.
        passages_per_prompt (int, optional): Number of paragraphs to include in each passage. Defaults to 4. Reduced if prompts are too long.
        step (int, optional): Overlap between passages. Defaults to 1.

    Raises:
        TooLongRequestError: If context is too long even after reducing the number of passages per prompt

    Returns:
        Sequence[PassageWithReasoning[float]]: Ranked passages in descending order, with reasoning and helpfulness for each passage.
    """
    try:
        candidates = window_dropping(items=passages, n=passages_per_prompt, step=step)
        assert best_choice in choices, "best_choice should be in choices"
        for choice in choices:
            if not choice[0].isspace():
                log.warn("Choice not starting with leading whitespace", choice=choice)

        async def score(
            candidates: Sequence[str],
        ) -> PassageWithReasoning[float]:
            return await answer_with_best_reasoning(
                num_samples=num_samples,
                selector=rank_passages_selector,
                num_examples=num_shots,
                texts=candidates,
                reasoning_temperature=0.4 if num_samples > 1 else 0.0,
                reasoning_stop=reasoning_stop,
                prompt_func=prompt_func,
                get_reasoning=get_reasoning,
                get_helpfulness=get_helpfulness,
                final_answer_processor=lambda resp: choices_log_probs(resp, choices)[
                    best_choice
                ],
            )

        answers = await map_async(
            candidates, score, max_concurrency=10, show_progress_bar=False
        )

        sorted_answers = sorted(
            answers,
            key=lambda prs: prs.final_answer
            if prs.final_answer is not None
            else float("-inf"),
            reverse=True,
        )
        return sorted_answers

    except TooLongRequestError as e:
        if passages_per_prompt > 1:
            passages_per_prompt -= 1
            log.warn(
                "Using fewer passages per prompt",
                passages_per_prompt=passages_per_prompt,
                error_detail=e.detail,
            )

            return await rate_helpfulness_with_reasoning(
                passages,
                prompt_func,
                choices,
                best_choice,
                num_samples=num_samples,
                reasoning_stop=reasoning_stop,
                get_reasoning=get_reasoning,
                get_helpfulness=get_helpfulness,
                num_shots=num_shots,
                passages_per_prompt=passages_per_prompt,
                step=step,
            )
        else:
            raise e


recipe.main(rate_helpfulness_with_reasoning)



from collections.abc import Mapping
from collections.abc import Sequence

from structlog.stdlib import get_logger

from ice.apis.openai import openai_complete
from ice.recipe import recipe
from ice.recipes.program_search.nodes.prune.prompts import EXAMPLE_SEPARATOR
from ice.recipes.program_search.nodes.prune.prompts import (
    get_pruned_selections_via_logprobs,
)
from ice.recipes.program_search.nodes.prune.prompts import make_pruning_prompt
from ice.recipes.program_search.nodes.prune.prompts import (
    make_pruning_with_reasoning_prompt,
)
from ice.recipes.program_search.types import remove_lowest_perplexity
from ice.utils import n_tokens

log = get_logger()


# class Select(Protocol):
#     async def __call__(self, question: str, texts: list[str], examples: list[Example]) -> list[int]:
#         pass


def last_token_logprob(openai_response: dict) -> float:
    return openai_response["choices"][0]["logprobs"]["token_logprobs"][-1]


def last_token_top_logprobs(openai_response: dict) -> dict[str, float]:
    return openai_response["choices"][0]["logprobs"]["top_logprobs"][-1]


def logprobs_greater_than_none(
    selections: Mapping[int, float], none_logprob: float, texts: Sequence[str]
) -> Sequence[str]:
    return [text for idx, text in enumerate(texts) if selections[idx] > none_logprob]


async def prune(question: str, texts: list[str], max_to_keep: int) -> list[str]:
    prompt = make_pruning_prompt(question=question, existing=texts)
    response = await openai_complete(
        prompt=prompt, max_tokens=80, logprobs=100, echo=False, stop=EXAMPLE_SEPARATOR
    )
    selection_probs = get_pruned_selections_via_logprobs(
        response["choices"][0]["logprobs"], num_selections=len(texts)
    )
    selections = sorted(selection_probs, key=selection_probs.__getitem__, reverse=True)[
        :max_to_keep
    ]
    return [texts[selection] for selection in selections]


async def prune_with_reasoning(
    question: str,
    texts_with_perplexities: Sequence[tuple[str, float]],
    max_to_keep: int,
) -> list[str]:
    # todo: compare first logprob method to completion parse method
    if max_to_keep > 5:
        log.warning(
            "the openai api only returns the top 5 logprobs, so we cannot keep more than 5 candidates via logprobs.",
            num_candidates=len(texts_with_perplexities),
        )
    prompt = make_pruning_with_reasoning_prompt(
        question=question, existing=[t[0] for t in texts_with_perplexities]
    )
    while n_tokens(prompt) > 3200:
        texts_with_perplexities = remove_lowest_perplexity(texts_with_perplexities)
        prompt = make_pruning_with_reasoning_prompt(
            question=question, existing=[t[0] for t in texts_with_perplexities]
        )
    response = await openai_complete(
        prompt=prompt,
        max_tokens=4000 - n_tokens(prompt),
        logprobs=100,
        echo=False,
        stop=EXAMPLE_SEPARATOR,
    )
    completion: str = response["choices"][0]["text"]
    if "from most to least important: " not in completion:
        log.warning("Unexpected completion", prompt=prompt, completion=completion)
        prompt = (
            prompt
            + completion.rstrip()
            + "\n\nWhich excerpts answer the question, from most to least"
        )
        response = await openai_complete(
            prompt=prompt,
            max_tokens=4090 - n_tokens(prompt),
            logprobs=100,
            echo=False,
            stop=EXAMPLE_SEPARATOR,
        )

    selection_probs = get_pruned_selections_via_logprobs(
        response["choices"][0]["logprobs"], num_selections=len(texts_with_perplexities)
    )
    selections = sorted(selection_probs, key=selection_probs.__getitem__, reverse=True)[
        :max_to_keep
    ]
    texts = [t[0] for t in texts_with_perplexities]
    return [texts[selection] for selection in selections]


recipe.main(prune)



from collections.abc import Callable
from collections.abc import Iterable
from functools import partial
from itertools import islice
from typing import Type

from ice.formatter.transform.value import numbered_list
from ice.metrics.gold_standards import generate_papers_and_golds
from ice.metrics.gold_standards import GoldStandard
from ice.metrics.gold_standards import ParsedGoldStandardType
from ice.paper import Paper
from ice.recipes.consort_flow.types import ConsortFlow
from ice.recipes.experiments_and_arms.types import ExperimentsArms
from ice.recipes.meta.eval_paper_qa.types import PaperQaGoldStandard


def generate_paper_qa_gold_standards(
    split: str,
    gold_standard_type: Type[ParsedGoldStandardType],
    question_and_answer_generator: Callable[
        [Paper, GoldStandard[ParsedGoldStandardType]],
        Iterable[PaperQaGoldStandard],
    ],
) -> Iterable[PaperQaGoldStandard]:
    for paper, gold in generate_papers_and_golds(split, gold_standard_type):
        yield from question_and_answer_generator(paper, gold)


def experiments_questions_and_answers(
    paper: Paper, gold: GoldStandard[ExperimentsArms], consolidate: bool = False
) -> Iterable[PaperQaGoldStandard]:
    if not gold.parsed_answer:
        return
    experiments = gold.parsed_answer.experiments
    gold_answer = [f"{exp.name}: {exp.description}" for exp in experiments]
    short_gold_answer = [exp.name for exp in experiments]
    if consolidate:
        gold_answer = (
            numbered_list(gold_answer).transform()
            + "\n\n"
            + f"({len(gold_answer)} experiment{'s' if len(gold_answer) > 1 else ''} in total)"
        )
        short_gold_answer = numbered_list(short_gold_answer, separator=", ").transform()

    question = """Experiments are distinct from trial arms or groups; a single experiment might have multiple trial arms, like different interventions or controls. What experiment or experiments (aka trials, RCTs, studies) were conducted in this paper? Enumerate them, being mindful that there may just be one experiment or there could be more than one. Include the name and a brief description of each experiment."""
    gold_support = gold.quotes
    yield PaperQaGoldStandard(
        paper=paper,
        question=question,
        gold_answer=gold_answer,
        short_gold_answer=short_gold_answer,
        gold_support=gold_support,
    )


def generate_experiments_qas_lst(split: str):
    yield from generate_paper_qa_gold_standards(
        split, ExperimentsArms, experiments_questions_and_answers
    )


def only_10_generate_experiments_qas_lst(split: str):
    yield from islice(
        generate_paper_qa_gold_standards(
            split, ExperimentsArms, experiments_questions_and_answers
        ),
        10,
    )


def generate_experiments_qas_str(split: str):
    yield from generate_paper_qa_gold_standards(
        split,
        ExperimentsArms,
        partial(experiments_questions_and_answers, consolidate=True),
    )


def arms_questions_and_answers(
    paper: Paper, gold: GoldStandard[ExperimentsArms], consolidate: bool = False
) -> Iterable[PaperQaGoldStandard]:
    if not gold.parsed_answer:
        return
    experiments = gold.parsed_answer.experiments
    all_exps = numbered_list(
        [f"{exp.name}: {exp.description}" for exp in experiments], separator=" / "
    ).transform()
    for experiment in experiments:
        gold_answer = [f"{arm.name}: {arm.description}" for arm in experiment.arms]
        short_gold_answer = [arm.name for arm in experiment.arms]
        if consolidate:
            gold_answer = (
                numbered_list(gold_answer).transform()
                + "\n\n"
                + f"({len(gold_answer)} arm{'s' if len(gold_answer) > 1 else ''} in total)"
            )
            short_gold_answer = numbered_list(
                short_gold_answer, separator=", "
            ).transform()
        if len(experiments) > 1:
            question = f"""This paper studied multiple experiments: {all_exps}. For the {experiment.name} experiment specifically, what were the different trial arms (subgroups of participants)?""".strip()
        else:
            question = f"What were the different trial arms (subgroups of participants) in the {experiment.name} ({experiment.description}) experiment?"
        yield PaperQaGoldStandard(
            paper=paper,
            question=question,
            gold_answer=gold_answer,
            short_gold_answer=short_gold_answer,
            gold_support=gold.quotes,
        )


def generate_arms_qas_lst(split: str):
    yield from generate_paper_qa_gold_standards(
        split, ExperimentsArms, arms_questions_and_answers
    )


def generate_arms_qas_str(split: str):
    yield from generate_paper_qa_gold_standards(
        split,
        ExperimentsArms,
        partial(arms_questions_and_answers, consolidate=True),
    )


def adherence_questions_and_answers(
    paper: Paper, gold: GoldStandard[ConsortFlow]
) -> Iterable[PaperQaGoldStandard]:
    if not gold.parsed_answer:
        return
    experiments = gold.parsed_answer.experiments
    for experiment in experiments:
        arms = experiment.arms
        if not arms:
            return
        for arm in arms:
            EXPLANATION = "Adherence describes how many participants selected for an intervention actually received it. Attrition describes how many of the initial sample dropped out of the study or were otherwise not available to be included in the final analysis. Compliance describes how well participants in the intervention complied with its protocol."
            question = f"""{EXPLANATION} For the {arm.name} arm of the {experiment.name} ({experiment.description}) experiment specifically, what does the paper say about the adherence, attrition, or compliance rate (say 'not mentioned' if it is not mentioned)?"""
            adherence = arm.received
            if not adherence:
                continue
            gold_answer = (
                adherence if isinstance(adherence, str) else adherence.description
            )
            if not gold_answer:
                continue
            yield PaperQaGoldStandard(
                paper=paper,
                question=question,
                gold_answer=gold_answer,
                short_gold_answer=gold_answer,
                gold_support=[] if isinstance(adherence, str) else adherence.quotes,
            )


def generate_adherence_qas(split: str):
    yield from generate_paper_qa_gold_standards(
        split, ConsortFlow, adherence_questions_and_answers
    )



from pathlib import Path

from fastapi import APIRouter
from fastapi import Header
from fastapi import HTTPException
from starlette.responses import FileResponse
from starlette.responses import PlainTextResponse

from ice.trace import traces_dir

router = APIRouter(prefix="/api/traces", tags=["traces"])


@router.get("/")
async def list_traces():
    # e.g. if trace_dir contains files trace1/trace.jsonl, trace2/trace.jsonl, other.txt,
    # return ["trace1", "trace2"]
    return [
        folder.name
        for folder in traces_dir.iterdir()
        if folder.is_dir() and (folder / "trace.jsonl").exists()
    ]


@router.get("/{trace_id}/trace.jsonl")
async def get_trace(trace_id: str, Range: str | None = Header(None)):
    """
    Return the contents of the trace file with the given trace_id.
    Uses the Range header to support partial content requests.
    This route comes before the StaticFiles mounted at /api/traces/ so it takes precedence.
    We need this because StaticFiles doesn't support Range headers.
    We still have the StaticFiles to support HEAD requests used for getting Content-length.
    """
    return get_file(traces_dir / trace_id / "trace.jsonl", Range)


@router.get("/{trace_id}/block_{block_id}.jsonl")
async def get_block(trace_id: str, block_id: int, Range: str | None = Header(None)):
    return get_file(traces_dir / trace_id / f"block_{block_id}.jsonl", Range)


def get_file(path: Path, Range: str | None):
    if not path.exists():
        raise HTTPException(status_code=404, detail="File not found")

    if Range is None:
        return FileResponse(path)

    try:
        start, end = map(int, Range.removeprefix("bytes=").split("-"))
    except ValueError:
        raise HTTPException(status_code=400, detail="Invalid Range header")

    with open(path, "rb") as f:
        f.seek(start)
        length = end - start + 1
        byts = f.read(length)
        text = byts.decode("utf-8", errors="ignore")
        return PlainTextResponse(text, status_code=206)



import typing as t

from pydantic import BaseModel
from pydantic.generics import GenericModel

from ice.paper import Paper


def _to_str(paper: Paper, start: int, end: int) -> str:
    return " ".join(list(paper.sentences())[start:end])


class Selection(BaseModel):
    p: Paper
    start: int
    end: int
    is_gs: bool | None = None

    # TODO: validate on creation

    def context(self, pre: int, post: int):
        idxs_before = range(max(0, self.start - pre), self.start)
        idxs_after = range(
            self.end, min(len(list(self.p.sentences())), self.end + post)
        )
        before = [
            Selection(p=self.p, start=start, end=start + 1) for start in idxs_before
        ]
        after = [Selection(p=self.p, start=end - 1, end=end) for end in idxs_after]
        return before, after

    @t.final
    @property
    def original(self):
        return _to_str(self.p, self.start, self.end)

    def __str__(self):
        return _to_str(self.p, self.start, self.end)

    class Config:
        copy_on_model_validation = False


def make_selector(paper: Paper) -> t.Callable[[int, int], Selection]:
    def s(start: int, end: int) -> Selection:
        return Selection.construct(p=paper, start=start, end=end)

    return s


def sentences(paper: Paper) -> t.Sequence[Selection]:
    s = make_selector(paper)
    return [s(start, start + 1) for start in range(0, len(list(paper.sentences())))]


def text_to_selection(text: str, selections: t.Sequence[Selection]):
    selection_dict = {selection.original: selection for selection in selections}
    return selection_dict[text]


class Decontext(Selection):
    questions: t.Sequence[str] | None
    out: str

    def __str__(self):
        return self.out


class Trace(BaseModel):
    components: t.Sequence[str | Selection]


T = t.TypeVar("T")


class Beam(GenericModel, t.Generic[T]):
    ...


def remove_lowest_perplexity(results: t.Sequence[tuple[str, float]]):
    drop = min(range(len(results)), key=lambda idx: results[idx][1])
    return list(results[0:drop]) + list(results[drop + 1 :])


def remove_highest_perplexity(results: t.Sequence[tuple[str, float]]):
    drop = max(range(len(results)), key=lambda idx: results[idx][1])
    return list(results[0:drop]) + list(results[drop + 1 :])
